/** \file DXModel.h
<h2> Model loader and renderer for Direct3D using the Assimp library </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Model loader and renderer
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#ifndef __DX_MODELS__
#define __DX_MODELS__

#include <map>
#include <vector>
#include <ai/assimp.hpp>
#include <ai/assimp.h>
#include <ai/aiPostProcess.h>
#include <ai/aiScene.h>
#include "DXInterface.h"
#include "DXMatrix.h"
#include "DXShader.h"
#include "DXQueue.h"

using namespace Assimp;
using namespace std;

#define MODEL_SCENEPATH_HEADERNUMBER -31337
#define MODEL_SCENEPATH_SEPARATOR -1234
#define MODEL_TECHNIQUE_BUFFERLEN 64

//! Model class - handles an engine model
class Model {
protected:

    //! Assimp scene handle
    const aiScene* scene;
    //! Assimp model importer class
    Importer modelImporter;

    //! Handle to current active model shader
    Shader *currentShader;
    //! Handle to current active model shader's semantics
    ShaderGlobalSemantics *currentShaderSem;
    
    //! Current model scene tree path
    vector<int> sceneTreePath;
    //! Model Direct3D vertexbuffer table
    map< vector<int>, PolygonQueue* > vertexBufferTable;

    //! Technique string buffer, used for materials
    char tempTechniqueBuffer[ MODEL_TECHNIQUE_BUFFERLEN ];


    //! Display the current scene tree path
    void printCurrentTreePath();

    //! Recurve function to traverse, read, build vertexbuffers and render the model in Direct3D
    void modelRender( const aiScene *scene, const aiNode *nd, const char *techniqueName = NULL, int depth = 0 );

public:

    Model();
    ~Model();

    //! Whether to render the model flipped around the X axis
    bool flipX;
    //! Whether to render the model flipped around the Y axis
    bool flipY;

    //! Whether to enable model materials
    bool materialsEnabled;
    //! When set to something above 0, ONLY enables rendering of
    //  parts of the model that given material ID.
    int  materialDebug;
    //! Optional callback function to freely modify material indexes
    (int)(*materialCallback)(int n);


    // ---> Public interface functions
    
    //! Load a model from file
    void load( const char* fileName );

    //! Clears the vertexbuffer table
    void clearVertexBufferCache();

    //! Renders the model
    void render(Shader *shader = NULL, ShaderGlobalSemantics *sgs = NULL, const char *techniqueName = NULL);
    
    //! Releases memory associated with the model
    void release();
};


#endif