/** \file DXPrimitives.h
<h2> Direct3D Primitives Rendering - Renders spheres, rings, cubes...etc </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Direct3D Primitives Rendering
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/


#ifndef __DX_PRIMITIVES__
#define __DX_PRIMITIVES__

#include <cmath>
#include "DXInterface.h"
#include "DXQueue.h"
#include <vector>

using namespace std;

//! Primitive object structure
struct dxPrimitivesObj {
    //! The Direct3D vertexbuffer storing the vertices of the primitive
    LPDIRECT3DVERTEXBUFFER9 vertexBuffer;
    //! Number of slices (optional)
    int slices;
    //! Number of stacks (optional)
    int stacks;
    //! Number of polygons in the vertex buffer
    int nPolygons;

    //! Releases the object's assigned vertex buffer memory
    void release();
};

//! Initialises the global primitive vertexbuffers and other info
void dxPrimitiveInit();

//! Releases the global primitive vertexbuffers
void dxPrimitiveRelease();

//! Renders a basic unit cube
void dxCube();

//! Renders a minimal unit quad
void dxFastQuad();

//! Renders a quad
void dxQuad( float width= 2.0f, float height = 2.0f, float xo = -1.0f, float yo = -1.0f, DWORD colour = 0xffffffff, bool RHW = false );

//! Renders a fullscreen picture
void dxBackgroundPicture(int width = d3dWindowWidth, int height = d3dWindowHeight, DWORD colour = 0xffffffff);

//! Renders a teapot
void dxTeapot();

//! Renders a sphere primitive object
void dxSphere( dxPrimitivesObj *meshSphere );

//! Renders a ring primitive object
void dxRing( dxPrimitivesObj *objRing ) ;

//! Renders a tube primitive object
void dxTube( dxPrimitivesObj *objTube );

//! Renders a torus primitive object
void dxTorus( dxPrimitivesObj *objTorus );

//! Creates a torus primitive object
dxPrimitivesObj *dxPrimitiveCreateTorus(float radius = 1.0f, float radius2 = 0.1f, int slices = 24, int sections = 24);

//! Creates a sphere primitive object
dxPrimitivesObj *dxPrimitiveCreateSphereMesh(float radius = 1.0f, unsigned int slices = 24, unsigned int stacks = 24);

//! Creates a ring primitive object
dxPrimitivesObj *dxPrimitiveCreateRing(float radius = 1.0f, float innerRadius = 0.5f, int slices = 24);

//! Creates a tube primitive object
dxPrimitivesObj *dxPrimitiveCreateTube(float radius = 1.0f, float length = 1.0f, int slices = 24, int sections = 2);

//! Calculates triangle tangent vector, given 3 UV-assigned vertexes
void dxPrimitiveTangent( DXVertex3 &v1, DXVertex3 &v2, DXVertex3 &v3);

//! Extrudes given polygon edge and puts result into givern polygon queue
void dxExtrude( DXVertex3 v0, DXVertex3 v1, float dis, PolygonQueue &pq );


#endif