/** \file DXRandom.h
<h2> Random / heuristic utility functions </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Random / heuristic utility functions
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#ifndef __DX_RANDOM_UTILITY__
#define __DX_RANDOM_UTILITY__

#include <cstdio>
#include <cmath>
#include <ctime>
#include <climits>

// ============================== Mathematical Constants ============================

#define _PI 3.14159265358979323846
#define _PI_OVER_TWO 1.57079632679489661923
#define _TWOPI 6.28318530717958647692
#define _E 2.718281828459045
#define _GOLDENRATIO 1.6180339887

// ============================== Time Hasher ============================
//! Hashes time - for use as seeding for RNGs
unsigned long _timeseed(void);

// ============================== Linear Congruential Generator ============================
//! Seeds linear congruential random number generator - works like c srand()
void _srand_linc( unsigned long seed = _timeseed() );
//! Linear congruential generator, given option to customise multiplier A and constant C
unsigned long _rand_linc_custom(unsigned long A, unsigned long C);
//! Linear congruential generator, using GCC constants
unsigned long _rand_linc(void);
//! Linear congruential generator, using Mininal Standard Generator constants
unsigned long _rand_linc_msg(void);
//! Linear congruential generator, using Borland C/C++ constants
unsigned long _rand_linc_borlandc(void);
//! Linear congruential generator, using Microsoft Visual C++ constants
unsigned long _rand_linc_msvc(void);

// ============================== XORShift ============================
//! XORShift Random Number Generator
unsigned long _rand_xor(void);

// ============================== Mersenne Twister ============================
//! Seeds Mersenne twister matrix - works like cstdlib srand()
void _srand( unsigned long seed = _timeseed() );
//! Used interally by Mersenne twister to fill matrix
void _gen(void);
//! Gives a (very) random number - works like cstdlib rand()
unsigned long _rand(void);
//! Gives a (very) random float, in a given range
float _randfloat(float min, float max);
//! Gives a (very) random number, in a given range
unsigned long _randrange(unsigned long mod);

// ============================== Noise Functions ============================
//! Noise function 1 - From Huge Elias' page
float _noise1(int x);
//! Noise function 2 - Variation of noise1
float _noise2(int x);
//! Noise function 2D - 2-Dimensional noise function
float _noise2D(int x, int y);

// ============================== Perlin Noise ============================
//! Smoothed noise by weighted avg - Used internally (mostly)
float _noiseSmooth(int x, int seed);
//! Does cosine interpolation between two values - Used internally (mostly)
float _interpolateCosine(float a, float b, float frac);
//! Calculates smooth interpolated noise - Used internally (mostly)
float _interpolateNoise(float x, int seed);
//! Perlin Noise Generator : 1-Dimensional 
float _perlinNoise(float x, int seed = 0, float persistence = 0.5f, int octaves = 4);

// ============================== General Purpose Hashing Functions ============================ 
unsigned int RSHash  (char* str, unsigned int len);
unsigned int JSHash  (char* str, unsigned int len);
unsigned int PJWHash (char* str, unsigned int len);
unsigned int ELFHash (char* str, unsigned int len);
unsigned int BKDRHash(char* str, unsigned int len);
unsigned int SDBMHash(char* str, unsigned int len);
unsigned int DJBHash (char* str, unsigned int len);
unsigned int DEKHash (char* str, unsigned int len);
unsigned int BPHash  (char* str, unsigned int len);
unsigned int FNVHash (char* str, unsigned int len);
unsigned int APHash  (char* str, unsigned int len);

// ============================== General Purpose Hashing Functions ============================ 
//! Jenkins Hash - Pretty good, apparently
unsigned int JenkinsHash(unsigned char* k, unsigned int len, unsigned int seed);

// ============================== Fast Float Functions ============================ 
//! Fast float absolute value
float _fabs(float f);
//! Fast float negation
float _fneg(float f);
//! Gives back +1 for 0 or positive numbers, -1 for negative numbers
int _fsign(float f);

// ============================== Fast Approximation Functions ============================ 
//! iD Software's famous inverse square root approximation
float _invsqrt (float x);

//! Fast sqrt using iD Software inverse sqrt
float _sqrt(float x);

//! Fast calculation of floor of log2
//! assert( f > 0. );
//! assert( sizeof(f) == sizeof(int) );
//! assert( sizeof(f) == 4 );
int _floor_log2( float f );

//! Fast approximation of log2
int _log2( float f );

//! Fast exp() approximation in the -3.14..3.14 range.
//! The increasing numbers in the name of the function means increasing precision. 
float fastexp3(float x);
float fastexp4(float x);
float fastexp5(float x);
float fastexp6(float x);
float fastexp7(float x);
float fastexp8(float x);
float fastexp9(float x);

//! Fast log^2 approximation
//!    assert (val > 0);
float _flog2(float val);

//! Fast pow() f^n approximation - rough estimate!
float _fpow(float f, int n);

//! Fast approximation for the n-th root of f.
float _froot(float f, int n);

// ============================== Fast Trigonometry Approximation Functions ============================ 
//! Extra fast sin() approximation : [0, pi/2]
float _sinfast(float fAngle);
//! Fast sin() approximation : [0, pi/2]
float _sin(float fAngle);
//! Extra fast cos() approximation : [0, pi/2]
float _cosfast(float fAngle);
//! Fast cos() approximation : [0, pi/2]
float _cos(float fAngle);
//! Extra fast tan() approximation : [0,pi/4]
float _tanfast(float fAngle);
//! Fast tan() approximation : [0,pi/4]
float _tan(float fAngle);
//! Fast inverse sin() approximation : [0, 1]
float _asin(float fValue);
//! Fast inverse cos() approximation : [0, 1]
float _acos(float fValue);
//! Extra fast inverse tan() approximation : [-1, 1]
float _atanfast(float fValue);
//! Fast inverse tan() approximation : [-1, 1]
float _atan(float fValue);

// ============================== Binomial Coefficients ============================ 
//! Calculates nCm using dynamics programming. n, m < 100
int _binomial_coefficient(int n, int m);

// ============================== Greatest Common Divisor ============================ 
//! Calculates gcd using Euclid's algorithm
int gcd(int a, int b);

// ============================== Integer Bithacks ============================ 
//! Branchless integer abs()
int _abs(int v);
//! Branchless min()
int _min(int x, int y);
//! Branchless max()
int _max(int x, int y);
//! Round up integer to next highest power of 2
unsigned int _nextpowerof2(unsigned int v);

// ============================== Gaussian Noise ============================ 
//! Gaussian noise random number generator - range 1 to -1. 
//! Uses Mersenne twister, so seed using _srand().
float _gaussianNoise();

#endif
