/** \file DXShadow.h
<h2> Shadow Map Rendering </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Shadow Map Rendering 
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#ifndef __DX_SHADOW__
#define __DX_SHADOW__

#include "DXInterface.h"
#include "DXRenderPass.h"
#include "DXShader.h"

#define DX_SHADOWMAP_DEFAULT_SIZE 512
#define DX_SHADOWMAP_DEFAULT_FORMAT D3DFMT_A8R8G8B8
#define DX_SHADOWMAP_DEFAULT_ZNEAR 0.1f
#define DX_SHADOWMAP_DEFAULT_ZFAR 10000.0f

//! Class that manages rendering from a light's POV to
//  a shadowmap
class ShadowMap {
private:
    D3DXMATRIX oldViewMatrix;
    D3DXMATRIX oldProjMatrix;

    D3DXMATRIX lightViewMatrix;
    D3DXMATRIX lightProjMatrix;

    D3DXVECTOR3 lightPos;
    D3DXVECTOR3 lightTarget;
    D3DXVECTOR3 lightUp;

    float zNear;
    float zFar;

    int shadowMapWidth;
    int shadowMapHeight;
    int shadowMapBoxWidth;
    int shadowMapBoxHeight;

protected:
    //! The shadowmap render buffer
    RenderPass shadowPass;

    //! Gets current Direct3D view and projection matrices
    void getViewProjection( D3DXMATRIX &viewMatrix, D3DXMATRIX &projMatrix );

public:

    //! Enables rendering to the shadowmap buffer (renders to screen otherwise)
    bool renderToShadowMap;
    //! Enables dumping rendered shadowmap to file
    bool dumpShadowMap;

    ShadowMap();
    ~ShadowMap();

    //! Creates shadow map buffer
    void initalise( D3DFORMAT format = DX_SHADOWMAP_DEFAULT_FORMAT, 
        int width = DX_SHADOWMAP_DEFAULT_SIZE, int height = DX_SHADOWMAP_DEFAULT_SIZE,
        int boxWidth = DX_SHADOWMAP_DEFAULT_SIZE, int boxHeight = DX_SHADOWMAP_DEFAULT_SIZE);

    //! Pushes on the light's POV state
    void pushState();

    //! Restores old view and projection matrices and states
    void popState();

    //! Set the position of the light
    void setLightPos( const D3DXVECTOR3 &p );

    //! Set light target directly
    void setLightTarget( const D3DXVECTOR3 &t );

    //! Set light target given direction (position has to be set first)
    void setLightDir( const D3DXVECTOR3 &d );

    //! Set light's up vector
    void setLightUp( const D3DXVECTOR3 &up );

    //! Put light's POV rendering matrices into sematics structure
    void setSemantics( ShaderGlobalSemantics *shaderSem );

    //! Get the light's current transformation matrix (WorldViewProj)
    D3DXMATRIX getLightMatrix( ShaderGlobalSemantics *shaderSem );

    //! Get the shadow map buffer texture
    operator LPDIRECT3DTEXTURE9();

    //! Release the buffers and handles associated
    void release();

};

#endif