/** \file DXMatrix.cpp
<h2> Direct3D Matrix Stack </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Direct3D Matrix Stack - handles world matrix states
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXMatrix.h"
#include <vector>

using namespace std;

//! The global matrix stack
vector<D3DXMATRIX> d3dMatrixStack;

//! Get a copy of the current world matrix
D3DXMATRIX dxGetMatrix() { 
    D3DXMATRIX i; D3DXMatrixIdentity( &i );
    return d3dMatrixStack.empty()?i:d3dMatrixStack[d3dMatrixStack.size()-1]; 
}

//! Set the stack to given matrix
void dxSetMatrix( D3DXMATRIX newMatrix ) { d3dMatrixStack[d3dMatrixStack.size()-1] = newMatrix; }

//! Returns a reference of the current world matrix
D3DXMATRIX *dxTopMatrix() { return d3dMatrixStack.empty()?NULL:&d3dMatrixStack[d3dMatrixStack.size()-1]; }

//! Push current matrix stack
void dxPushMatrix() {
    D3DXMATRIX newMatrix = dxGetMatrix();
    d3dMatrixStack.push_back(newMatrix);
}

//! Pop current matrix stack
void dxPopMatrix() {
    if( !d3dMatrixStack.empty() ) d3dMatrixStack.pop_back();
    D3DXMATRIX newMatrix = dxGetMatrix();
}

//! Push new matrix stack
void dxPushSetMatrix( D3DXMATRIX newMatrix ) {
    d3dMatrixStack.push_back(newMatrix);
}

//! Rotate current matrix stack top around X axis
void dxRotateX( float deg ) {
    D3DXMATRIX rot; D3DXMatrixRotationX(&rot, D3DXToRadian(deg));
    (*dxTopMatrix()) = rot * (*dxTopMatrix());
}

//! Rotate current matrix stack top around Y axis
void dxRotateY( float deg ) {
    D3DXMATRIX rot; D3DXMatrixRotationY(&rot, D3DXToRadian(deg));
    (*dxTopMatrix()) = rot * (*dxTopMatrix());
}

//! Rotate current matrix stack top around Z axis
void dxRotateZ( float deg ) {
    D3DXMATRIX rot; D3DXMatrixRotationZ(&rot, D3DXToRadian(deg));
    (*dxTopMatrix()) = rot * (*dxTopMatrix());
}

//! Translate current matrix stack top
void dxTranslate( float x, float y, float z ) {
    D3DXMATRIX tr; D3DXMatrixTranslation(&tr, x, y, z);
    (*dxTopMatrix()) = tr * (*dxTopMatrix());
}

//! Scale current matrix stack top
void dxScale(float sx, float sy, float sz) {
    D3DXMATRIX sc; D3DXMatrixScaling(&sc, sx, sy, sz);
    (*dxTopMatrix()) = sc * (*dxTopMatrix());
}

//! Multiply current matrix stack top by givern matrix
void dxMultiply( D3DXMATRIX &m ) {
    (*dxTopMatrix()) =  m *(*dxTopMatrix());
}

//! Update Direct3D with the current matrix stack top
void dxCommitMatrix() {
    if( !d3dMatrixStack.empty() ) d3dDevice->SetTransform(D3DTS_WORLD, dxTopMatrix());
}