/** \file DXRenderPass.cpp
<h2> Handles Rendering to Off-screen Buffer </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Handles Rendering to Off-screen Buffer
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXRenderPass.h"

// --------> RenderPass Methods

RenderPass::RenderPass() {
    targetWidth = 0; targetHeight = 0;
    for(int a = 0; a < MAX_POSTPROCESS_RENDERTARGETS; a++) {
        targetTexture[a] = NULL;
        targetTextureSurface[a] = NULL;
    }
}

RenderPass::~RenderPass() {
}

//! Creates a new render buffer using given dimensions and format
void RenderPass::createTarget(D3DFORMAT format, int width , int height, int index) {
    // Set up state variables
    this->releaseTarget();
    targetWidth = width; targetHeight = height;
    // Create target texture
    HRESULT hr = d3dDevice->CreateTexture( targetWidth, targetHeight, 1, D3DUSAGE_RENDERTARGET,
        format, D3DPOOL_DEFAULT, &targetTexture[index], NULL );
    if(hr != D3D_OK) {
        switch(hr) { 
            case D3DERR_INVALIDCALL: wError( "PostProcessPass::createTarget() CreateTexture() error: INVALIDCALL!"); return;
            case D3DERR_OUTOFVIDEOMEMORY: wError( "PostProcessPass::createTarget() CreateTexture() error: OUTOFVIDEOMEMORY!"); return;
            case E_OUTOFMEMORY: wError( "PostProcessPass::createTarget() CreateTexture() error: E_OUTOFMEMORY!"); return;
            default: wError( "PostProcessPass::createTarget() CreateTexture() error: unknown!"); return;
        }
    }
    // Create target Z stencil texture
    hr = d3dDevice->CreateTexture( targetWidth, targetHeight, 1, D3DUSAGE_DEPTHSTENCIL,
        DEPTH_STENCIL_FORMAT, D3DPOOL_DEFAULT, &targetZTexture[index], NULL );
    if(hr != D3D_OK) {
        switch(hr) { 
            case D3DERR_INVALIDCALL: wError( "PostProcessPass::createTarget() CreateTexture() error: INVALIDCALL!"); return;
            case D3DERR_OUTOFVIDEOMEMORY: wError( "PostProcessPass::createTarget() CreateTexture() error: OUTOFVIDEOMEMORY!"); return;
            case E_OUTOFMEMORY: wError( "PostProcessPass::createTarget() CreateTexture() error: E_OUTOFMEMORY!"); return;
            default: wError( "PostProcessPass::createTarget() CreateTexture() error: unknown!"); return;
        }
    }
    // Set target surface
    hr = targetTexture[index]->GetSurfaceLevel(0, &targetTextureSurface[index]);
    if(hr != D3D_OK) { wError( "PostProcessPass::createTarget() GetSurfaceLevel() error!"); return; }
    // Set target Z stencil surface
    hr = targetZTexture[index]->GetSurfaceLevel(0, &targetZTextureSurface[index]);
    if(hr != D3D_OK) { wError( "PostProcessPass::createTarget() GetSurfaceLevel() error!"); return; }
}

//! Releases all associated memory and render targets
void RenderPass::releaseTarget() {
    for(int a = 0; a < MAX_POSTPROCESS_RENDERTARGETS; a++) {
        if( targetTexture[a] ) targetTexture[a]->Release();
        if( targetTextureSurface[a] ) targetTextureSurface[a]->Release();
        if( targetZTexture[a] ) targetZTexture[a]->Release();
        if( targetZTextureSurface[a] ) targetZTextureSurface[a]->Release();
    }
}

//! Sets Direct3D to render to this buffer
void RenderPass::setAsTarget( int screenIndex, int index ) {
    if( index < 0 || index >= MAX_POSTPROCESS_RENDERTARGETS ) {
        wError( "PostProcessPass::setAsTarget() index out of bounds!");
        return;
    }

    d3dDevice->GetRenderTarget( screenIndex, &oldRenderTarget );
    d3dDevice->GetDepthStencilSurface( &oldRenderZTarget );
    d3dDevice->GetViewport( &oldViewport );
    D3DVIEWPORT9 newViewPort = oldViewport;
    newViewPort.Width = targetWidth;
    newViewPort.Height = targetHeight;
    d3dDevice->SetViewport( &newViewPort );

    HRESULT hr = d3dDevice->SetRenderTarget(screenIndex, targetTextureSurface[index]);
    if(hr != D3D_OK) { wError( "PostProcessPass::setAsTarget() SetRenderTarget() error!"); return; }

    hr = d3dDevice->SetDepthStencilSurface(targetZTextureSurface[index]);
    if(hr != D3D_OK) { wError( "PostProcessPass::setAsTarget() SetDepthStencilSurface() error!"); return; }

}

//! Restores Direct3D rendering target to saved previous state
void RenderPass::restoreTarget( int screenIndex ) {
    d3dDevice->SetViewport( &oldViewport );
    d3dDevice->SetRenderTarget(screenIndex, oldRenderTarget);
    d3dDevice->SetDepthStencilSurface(oldRenderZTarget);
}

//! Gets off-screen buffer texture given index
LPDIRECT3DTEXTURE9 RenderPass::getTargetTexture(int index) {
    if( index < 0 || index >= MAX_POSTPROCESS_RENDERTARGETS ) {
        wError( "PostProcessPass::getSourceTexture() index out of bounds!");
        return NULL;
    }
    if( !targetTexture[ index ] ) {
        wError( "PostProcessPass::getSourceTexture() NULL texture!");
    }
    return targetTexture[ index ];
}

//! Gets off-screen buffer texture given index
LPDIRECT3DTEXTURE9 RenderPass::operator [] (int index) {
    return this->getTargetTexture( index );
}

//! Gets off-screen buffer texture of index 0
RenderPass::operator LPDIRECT3DTEXTURE9() {
    return this->getTargetTexture( 0 );
}

//! Renders a buffer ( texture NOT automatically set, use dxTexture() )
void RenderPass::render() {
    dxBackgroundPicture(d3dWindowWidth, d3dWindowHeight);
}

//! Dump a screen buffer to a file in BMP format
void RenderPass::dump( const char* fileName, int index ) {
    D3DXSaveTextureToFile(fileName, D3DXIFF_BMP, getTargetTexture(index), NULL);
}

