/** \file DXSound.cpp
<h2> Sound functionality using FMOD library </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Sound & DSP functions
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXSound.h"

// --------> SoundStream Methods

SoundStream::SoundStream() {
    playing = false; ptr = NULL;
    channel = -1; used = false;
    vol = DXSTREAM_DEFAULT_VOL;
}

//! Open a stream from file
void SoundStream::open(char *fileName) {
    ptr = FSOUND_Stream_Open(fileName, FSOUND_LOOP_OFF, 0, 0);
    if( ptr ) used = true;
    else wError("SoundStream::open() failure!");
}

//! Stop playback of stream
void SoundStream::stop() {
    if( !ptr ) return;
    FSOUND_Stream_Stop( ptr ); 
    playing = false;
}

//! Free stream handle and memory
void SoundStream::free() {
    stop(); 
    FSOUND_Stream_Close( ptr );
}

//! Set volume of playing stream
void SoundStream::setVol(int vol) {
    if( playing && channel!=-1 )
        FSOUND_SetVolume( channel, vol ); 
    this->vol = vol;
}

//! Set frequency of playing stream
void SoundStream::setFreq(int freq) {
    if( playing && channel!=-1 )
        FSOUND_SetFrequency( channel, freq );
}

//! Start playback of stream
void SoundStream::play() {
    if( !playing && ptr ) {
        channel = FSOUND_Stream_Play( FSOUND_FREE, ptr );
        this->setVol( vol );
        playing = ( channel!=-1 );
    }
}

//! Seek to a time in milliseconds into playing stream
void SoundStream::seek(int ms) {
    if( ptr ) FSOUND_Stream_SetTime( ptr, ms );
}

// --------> Functions

//! Initalise the sound system
void dxSoundSystemInit() {
    FSOUND_Init ( 44100, 16, 0);
}

//! Activate or De-activate FFT module
void dxSoundSystemSetFFT( bool active ) {
    FSOUND_DSPUNIT *fmodFFTUnit = FSOUND_DSP_GetFFTUnit();
    FSOUND_DSP_SetActive( fmodFFTUnit, active );
}

//! Get current audio spectrum as a 512 float array
//  Activate FFT module using dxSoundSystemSetFFT() first
float *dxSoundSystemSpectrum() {
    return FSOUND_DSP_GetSpectrum();
}

//! Get current audio spectrum as low, mid, high levels.
//  Activate FFT module using dxSoundSystemSetFFT() first
void dxSoundSystemLMH( float *low, float *mid, float *high ) {
    (*low) = (*mid) = (*high) = 0.0f;
    int nlow = 0, nmid = 0, nhigh = 0;
    float *spectrum = dxSoundSystemSpectrum();
    for(int a = 0;a < DXSOUNDDSP_BANDS; a++) {
        float freq = (float)((float)a/(float)DXSOUNDDSP_BANDS) * (float)DXSTREAM_NYQUIST;
        if( freq <= DXSOUNDDSP_LOW ) {
            (*low)+= spectrum[a];
            nlow++;
        } else if ( freq >= DXSOUNDDSP_HIGH ) {
            (*high)+= spectrum[a];
            nhigh++;
        } else {
            (*mid)+= spectrum[a];
            nmid++;
        }
    }
    if( nlow > 0 ) (*low) /= (float)nlow;
    if( nmid > 0 ) (*mid) /= (float)nmid;
    if( nhigh > 0 ) (*high)/= (float)nhigh;
}
