/** \file DXTerrain.cpp
<h2> Height mapped terrain </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Height mapped terrain
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../EngineHeader/DXTerrain.h"

Terrain *currentSelectedTerrain;

// --------> Texture Methods

Terrain::Terrain() {
    gridDataSize = DXTERRAIN_DEFAULT_DSIZE;
    gridCellAspect = 1.0f;
    gridDimension = DXTERRAIN_DEFAULT_DIM;
    gridHeightScale = 1.0f / DXTERRAIN_HEIGHT_SCALE;
    gridCenter = D3DXVECTOR3(0.0f, 0.0f, 0.0f);
    miniHandle = NULL;
    elevationfunction = NULL;
    vertexBuffer = NULL;
    indexBuffer = NULL;
    vertexArray = NULL;
    indexArray = NULL;
    vertexArrayN = indexArrayN = fanCount = 0;
    fanLastVertex = fanBaseVertex = -1;
    emptyView = true;
    gridSize = 0.0f;
}

void Terrain::initalise() {
    release();
    miniHandle = new ministub( NULL, &gridDataSize, &gridDimension, gridHeightScale, gridCellAspect,
        gridCenter.x, gridCenter.y, gridCenter.z, 
        dxTerrainRenderBegin, dxTerrainRenderVertex, NULL, elevationfunction, NULL );
    vertexArray = new DXVertex3[ DXTERRAIN_MAX_VERTEX ];
    indexArray = new short[ 3 * DXTERRAIN_MAX_INDEX ];
    gridSize = (float)gridDataSize * gridDimension;
}

void Terrain::release() {
    if(miniHandle) delete miniHandle;
    if(vertexBuffer) {vertexBuffer->Release(); vertexBuffer = NULL;}
    if(indexBuffer) {indexBuffer->Release(); indexBuffer = NULL;}
    if(vertexArray) delete[] vertexArray;
    if(indexArray) delete[] indexArray;
}

void Terrain::compile(D3DXVECTOR3 cam, D3DXVECTOR3 view, D3DXVECTOR3 up, float resolution, float FOV, 
                     float aspect, float nearp, float farp) {
    if(!miniHandle) { wError("Terrain::compile() NULL miniHandle!"); return; }
    if(vertexBuffer) {vertexBuffer->Release(); vertexBuffer = NULL;}
    if(indexBuffer) {indexBuffer->Release(); indexBuffer = NULL;}
    // Set global target for static callback functions
    currentSelectedTerrain = this;
    fanCount = 0; fanLastVertex = fanBaseVertex = -1;
    vertexArrayN = indexArrayN = 0;
    emptyView = false;
    // Call the libmini render
    miniHandle->draw(resolution, cam.x, cam.y,-cam.z, view.x, view.y,-view.z, up.x, up.y,-up.z, FOV, aspect, nearp, farp);
    // Set vertex declaration
    if( vertexArrayN <= 0 || indexArrayN <= 0 ) { emptyView = true; return; }
    d3dDevice->SetVertexDeclaration( _DXVERTEXDECL3 );
    // Build vertex buffer
    HRESULT hr = d3dDevice->CreateVertexBuffer(vertexArrayN * sizeof(DXVertex3), 0, 0, D3DPOOL_MANAGED, &vertexBuffer, NULL);
    if(!vertexBuffer) { wError("Terrain::compile() vertexbuffer create failure!"); }

    // Lock and load vertex buffer
    DXVertex3 *dxVertices = NULL;
    vertexBuffer->Lock(0, 0, (void**)&dxVertices, 0);
    if(!dxVertices) { wError("Terrain::compile() vertexbuffer lock failure!"); return; }
    memcpy( dxVertices, vertexArray, vertexArrayN * sizeof(DXVertex3) );
    vertexBuffer->Unlock();
    // Build index buffer
    d3dDevice->CreateIndexBuffer( 3 * indexArrayN * sizeof(short), 0, D3DFMT_INDEX16, D3DPOOL_MANAGED, &indexBuffer, NULL);
    if(!indexBuffer) { wError("Terrain::compile() indexbuffer create failure!"); return; }
    // Lock and load index buffer
    short *dxIndices = NULL;
    indexBuffer->Lock( 0, 0, (void**)&dxIndices, 0);
    if(!dxIndices) { wError("Terrain::compile() indexbuffer lock failure!"); return; }
    memcpy(dxIndices, indexArray, 3 * indexArrayN * sizeof(short));
    indexBuffer->Unlock();
    // Clean up
    currentSelectedTerrain = NULL;
}

void Terrain::render() {
    if( emptyView ) return;
    if(!miniHandle) { wError("Terrain::render() NULL miniHandle!"); return; }
    if(!vertexBuffer) { wError("Terrain::render() NULL vertexBuffer!"); return; }
    if(!indexBuffer) { wError("Terrain::render() NULL indexBuffer!"); return; }
    dxPushMatrix();
        // Set vertex declaration
        d3dDevice->SetVertexDeclaration( _DXVERTEXDECL3 );
        // Render geometry
        d3dDevice->SetStreamSource(0, vertexBuffer, 0, sizeof(DXVertex3));
        d3dDevice->SetIndices(indexBuffer);
        d3dDevice->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, 0, vertexArrayN, 0, indexArrayN); 
    dxPopMatrix();
}

int Terrain::fanAddVertex( D3DXVECTOR3 pos ) {
    if(!vertexArray) return -1;
    if( vertexArrayN >= DXTERRAIN_MAX_VERTEX ) {
        wError("Terrain::fanAddVertex() - vertexArrayN overflow (terrain mesh too dense?)");
        return 0;
    }
    vertexArray[vertexArrayN].reset();
    vertexArray[vertexArrayN].pos = pos;
    vertexArray[vertexArrayN].setAllUV( pos.x, pos.y );
    return vertexArrayN++;
}

void Terrain::fanAddIndex( int r, int s, int t ) {
    if(r<0||s<0||t<0||r>=DXTERRAIN_MAX_INDEX||
        s>=DXTERRAIN_MAX_INDEX||t>=DXTERRAIN_MAX_INDEX) return;
    int i = indexArrayN * 3;
    indexArray[i  ] = r;
    indexArray[i+1] = s;
    indexArray[i+2] = t;
    indexArrayN++;
}

// --------> Functions

void dxTerrainRenderBegin() {
    if(!currentSelectedTerrain) { wError("dxTerrainRenderBegin() NULL pointer!"); return; }
    Terrain *tr = currentSelectedTerrain;
    // Start a new fan
    if(tr->fanCount == 0 ) printf("======================================\n");
    //printf("-------------- new Fan %d\n", tr->fanCount); 
    tr->fanLastVertex = tr->fanBaseVertex = -1;
    tr->fanCount++;
}

void dxTerrainRenderVertex(float i, float y, float j) {
    if(!currentSelectedTerrain) { wError("dxTerrainRenderVertex() NULL pointer!"); return; }
    Terrain *tr = currentSelectedTerrain;
    //printf("%f %f %f\n", i, y, j);
    i *= tr->gridDimension;
    j *= tr->gridDimension;
    y *= tr->gridHeightScale;
    if( tr->fanBaseVertex < 0 ) {
        tr->fanBaseVertex = tr->fanAddVertex( D3DXVECTOR3( i, y, j ) );
    } else if( tr->fanLastVertex < 0 ) {
        tr->fanLastVertex = tr->fanAddVertex( D3DXVECTOR3( i, y, j ) );
    } else {
        int newVertex = tr->fanAddVertex( D3DXVECTOR3( i, y, j ) );
        tr->fanAddIndex( tr->fanBaseVertex, tr->fanLastVertex, newVertex );
        tr->fanLastVertex = newVertex;
    }
}

