/** \file Render.cpp
<h2> Main rendering and initalisation module </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Main rendering and initalisation module
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../GameHeader/Render.h"
#include "../GameHeader/Planet.h"
#include "../EngineHeader/DXInfo.h"

#define MUSIC_BPM 134.1f

// --------------------------------------------------- Variables ----------------------------------------------------

// -----> Global Settings
int desiredScreenWidth = 1024;
int desiredScreenHeight = 768;
int desiredAniso = 0;

// -----> Camera controlling variables
int cameraControlMode = CAMERACONTROL_FREELOOK;
int cameraMouseDeltaX = 0, cameraMouseDeltaY = 0;
int cameraMouseLastX = -1, cameraMouseLastY = -1;
float cameraAngle1 = 0.0f, cameraAngle2 = 0.0f; 
D3DXVECTOR3 cameraLook, cameraUp, cameraTarget, cameraPos;
D3DXVECTOR3 freeViewCam = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
float cameraDistance = 1.0f;
float cameraMouseSensitivity = 0.1f;
float cameraDisplacementSpeed = 0.01f;
float cameraDistScrollSpeed = 5.0f;

// -----> Demo rendering state variables
bool trackPathMode = false;
bool spaceLoaded = false;
int musicState = -1;
int editMode = false;
int showFPS = true;
int logToFile = false;
float logFileTimer = 1000.0f;

// -----> Global textures array
TextureArray texture;

// -----> Demo scenes
SceneSolarSystem sceneSolarSystem;
SceneTerrain sceneTerrain;
SceneCubeTunnel sceneCubeTunnel;
SceneChrono sceneChrono;
SceneAbstract sceneAbstract;
SceneAbstract2 sceneAbstract2;
SceneHuman sceneHuman;
SceneCar sceneCar;

// -----> Current selected scene
Scene *sceneCurrent = NULL;

// -----> Global shader semantics
ShaderGlobalSemantics shaderSem;

// -----> Post processing variables
int HDREnabled = true;
int MLAAEnabled = true;
ShaderGlobalSemantics shaderSemDefault;
RenderPass screenHDRBuffer;
RenderPass screenBlurBufferU;
RenderPass screenBlurBufferV;
RenderPass screenRGBBuffer;
RenderPass screenRGBBuffer2;
RenderPass screenMLAA0Buffer;
RenderPass screenMLAA1Buffer;
Shader adjustmentPostProcess;
Shader blurPostProcess;
Shader MLAAPostProcess;

// -----> Post processing states
float postSaturation = 1.0f;
float postContrast = 1.0f;
float postExposure = 1.0f;
float postBlurAmount = 0.0f;
float postBlurAmount2 = 0.0f;
float postBloomAmount = 0.5f;
float postBloomThreshold = 0.7f;
float postWhite = 0.0f;
float postBlack = 1.0f;
float postNoise = 0.0f;
float postDispersion = 0.0f;
FadeController whiteFader;
FadeController blackFader;
FadeController noiseFader;
FadeController dispersionFader;
FadeController blurFader;

// -----> Animation path editing variables
CameraPath cameraPathEditor;
ParamPath paramPathEditor;

// -----> Music variables
SoundStream musicStream;
SoundStream musicStream2;

// --------------------------------------------------- Functions ----------------------------------------------------

//! Render the current console text
void renderConsole() {
    dxPrintf( 40, 30.0, 0xffffffff, "--------------------------------------------------------------");
    for(int a = 0 ;a < dxConsole.size(); a++) {
        dxPrintf( 40, 50 + 20.0 * a, 0xffffffff, "%s", dxConsole[a].c_str());
    }
}

//! Render the loading screen
void renderLoadingProgressFrame(const char* msg) {
    dxLock();
    dxTexture( 0, texture[1] );
    dxBackgroundPicture();
    dxPrintf( 40, 10, 0xffffffff, "%s", msg);
    renderConsole();
    dxUnLock();
    dxSwapBuffers();
}

//! Load settings from file
void loadSettings() {
    char tempBuffer[ CONFIG_BUFFER_SIZE ];
    FILE *a_file = NULL; fopen_s( &a_file, CONFIG_FILE, "r" );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &desiredScreenWidth );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &desiredScreenHeight );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &d3dWindowFullscreen );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &desiredAniso );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &MLAAEnabled );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &HDREnabled );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &editMode );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &showFPS );
    fscanf_s( a_file, "%s %d", tempBuffer, CONFIG_BUFFER_SIZE, &logToFile );
}

//! Loads all the demo scenes (except the space scene)
void loadScenes() {
    renderLoadingProgressFrame("Loading Elevated scene...");
    sceneTerrain.load();
    renderLoadingProgressFrame("Loading CubeTunnel scene...");
    sceneCubeTunnel.load();
    renderLoadingProgressFrame("Loading Chronosphere scene...");
    sceneChrono.load();
    renderLoadingProgressFrame("Loading Abstract scene...");
    sceneAbstract.load();
    renderLoadingProgressFrame("Loading Abstract2 scene...");
    sceneAbstract2.load();
    renderLoadingProgressFrame("Loading Human Head scene...");
    sceneHuman.load();
    renderLoadingProgressFrame("Loading Car scene...");
    sceneCar.load();
}

//! Unloads all the demo scenes (except the space scene)
void unloadScenes() {
    renderLoadingProgressFrame("Releasing Elevated scene...");
    sceneTerrain.release();
    renderLoadingProgressFrame("Releasing CubeTunnel scene...");
    sceneCubeTunnel.release();
    renderLoadingProgressFrame("Releasing Chronosphere scene...");
    sceneChrono.release();
    renderLoadingProgressFrame("Releasing Abstract scene...");
    sceneAbstract.release();
    renderLoadingProgressFrame("Releasing Abstract2 scene...");
    sceneAbstract2.release();
    renderLoadingProgressFrame("Releasing Human Head scene...");
    sceneHuman.release();
    renderLoadingProgressFrame("Releasing Car scene...");
    sceneCar.release();
}

//! Loads the space scene after unloading all other scenes
void loadSpaceScene() {
    if( spaceLoaded ) return;
    screenHDRBuffer.restoreTarget();
    // --> Unload everything else
    unloadScenes();
    // --> Load space scene
    renderLoadingProgressFrame("Loading Solar System scene...");
    sceneSolarSystem.load();
    spaceLoaded = true;
    // --> Reset timer
    wTickTime();
    execTime = dxTimerBPM( 134.1, 134, 0 );
    screenHDRBuffer.setAsTarget();
}

//! Appends current framerate to log file
void logCurrentFPS() {
    FILE *a_file = NULL; fopen_s( &a_file, "logFPS.txt", "a" );
    if( !a_file ) return;
    fprintf_s( a_file, "%.2f\n", frameRate );
    fclose( a_file );
}

//! Clears log file
void logReset() {
    FILE *a_file = NULL; fopen_s( &a_file, "logFPS.txt", "w" );
    if( a_file ) fclose( a_file );
}

//! Initialize DirectX, demo data and load contents
void initialize() {
    // Set random seed
    srand(_timeseed());
    _srand(_timeseed());
    // Dump system information and capabilities to console
    dxDisplayInfo();
    // Reset log file
    logReset();

    // Set Direct3D rendering states
    dxRenderState(D3DRS_LIGHTING, FALSE);
    dxRenderState(D3DRS_ZENABLE, D3DZB_TRUE);
    dxRenderState(D3DRS_NORMALIZENORMALS, TRUE);
    dxRenderState(D3DRS_AMBIENT, RGB(255,255,255));
    // Initalise primitive rendering 
    dxPrimitiveInit();
    // Initalise sound system
    dxSoundSystemInit();
    dxSoundSystemSetFFT( true );

    // Load shared textures
    texture.load(1, "Common/loading.png");
    texture.load(2, "Common/table.png");
    texture.load(3, "Common/noise.bmp");

    // Create post-processing buffers
    screenHDRBuffer.createTarget( HDREnabled?D3DFMT_A16B16G16R16F:D3DFMT_A8R8G8B8 );
    screenBlurBufferU.createTarget( HDREnabled?D3DFMT_A16B16G16R16F:D3DFMT_A8R8G8B8 );
    screenBlurBufferV.createTarget( HDREnabled?D3DFMT_A16B16G16R16F:D3DFMT_A8R8G8B8 );
    screenRGBBuffer.createTarget();
    screenRGBBuffer2.createTarget();
    screenMLAA0Buffer.createTarget();
    screenMLAA1Buffer.createTarget();

    // Enable / Disable Direct3D anisotropic filtering
    for(int a = 0; a < 8; a++) {
        dxSamplerState(a, D3DSAMP_MAXANISOTROPY, desiredAniso);
        dxSamplerState(a, D3DSAMP_MINFILTER, desiredAniso>0?D3DTEXF_ANISOTROPIC:D3DTEXF_LINEAR);
        dxSamplerState(a, D3DSAMP_MAGFILTER, desiredAniso>0?D3DTEXF_ANISOTROPIC:D3DTEXF_LINEAR);
        dxSamplerState(a, D3DSAMP_MIPFILTER, desiredAniso>0?D3DTEXF_ANISOTROPIC:D3DTEXF_LINEAR);
    }
 
    // Load scenes
    loadScenes();
    
    // Load music files
    musicStream.open( MUSIC_FILE );
    musicStream2.open( MUSIC_FILE2 );

    // Set up post-processing effects
    D3DXMatrixIdentity(&shaderSemDefault.matWorld);
    D3DXMatrixIdentity(&shaderSemDefault.matView);
    D3DXMatrixIdentity(&shaderSemDefault.matProj);
    shaderSemDefault.cameraPos = D3DXVECTOR3( 0.0f, 0.0f, 0.0f );
    adjustmentPostProcess.loadFile("Common/adjustment.fx");
    blurPostProcess.loadFile("Common/blur.fx");
    MLAAPostProcess.loadFile("Common/MLAA.fx");

    // Set up animation files and faders
    whiteFader.load("Common/whiteFader.txt");
    blackFader.load("Common/blackFader.txt");
    noiseFader.load("Common/noiseFader.txt");
    dispersionFader.load("Common/dispersionFader.txt");
    blurFader.load("Common/blurFader.txt");
}

//! Zoom in/out on mousewheel scroll
void wiMouseWheel( int delta ) {
    if( editMode ) cameraDistance += cameraDistScrollSpeed * delta;
}

//! Camera control on mouse click
void wiMouseMove( int x, int y ) {
    if( !editMode )  return;
    if( !rightMouse ) {
        cameraMouseLastX = cameraMouseLastY = -1;
        return;
    }
    if( cameraMouseLastX >= 0 && cameraMouseLastY >= 0 ) {
        cameraMouseDeltaX = x - cameraMouseLastX;
        cameraMouseDeltaY = y - cameraMouseLastY;
        cameraAngle1 += cameraMouseDeltaX * cameraMouseSensitivity;
        cameraAngle2 -= cameraMouseDeltaY * cameraMouseSensitivity;
    }
    cameraMouseLastX = x; cameraMouseLastY = y;
}

//! Key press handler - controls camera, control camera path editing,
//  and toggles / tweaks options
void wiKey( int key ) {
    double resetTime = 505.0f;
    double musicOffset = dxTimerBPM( MUSIC_BPM, 0, 0 );
    if( editMode ) {
        switch( key ) {
            case 'B': 
                wInfo( "> Adding camera frame %d...\n", cameraPathEditor.numFrames() );
                cameraPathEditor.addFrame( cameraPos, cameraTarget, cameraUp );
                if( sceneCurrent ) { sceneCurrent->addParamVector(&paramPathEditor); }
                break;
            case 'O':
                wInfo("> Saving current camera path...");
                cameraPathEditor.save("cameraPathOutput.txt");
                if( paramPathEditor.numFrames() > 0 )
                    paramPathEditor.save("paramPathOutput.txt");
                break;
            case '1': cameraControlMode = CAMERACONTROL_FREELOOK; break;
            case '2': cameraControlMode = CAMERACONTROL_FREEORBIT; break;
            case '3': cameraControlMode = CAMERACONTROL_PATH; execTime = 0.0; break;
            case '4': case '5': 
                cameraControlMode = CAMERACONTROL_DEMO; 
                execTime = (key == '5')?0.0:resetTime; 
                musicState = -1;        
                break;
            case 'X': cameraDisplacementSpeed *= 2.0f; cameraDistScrollSpeed *= 2.0f; break;
            case 'Z': cameraDisplacementSpeed /= 2.0f; cameraDistScrollSpeed /= 2.0f; break;
            case VK_SPACE: trackPathMode = !trackPathMode; break;
            case 'M': MLAAEnabled = !MLAAEnabled; break;
        }
    } else {
        switch( key ) {
            case 'M': MLAAEnabled = !MLAAEnabled; break;
        }
    }
}

//! End the demo and exit program
void endDemo() {
    if( spaceLoaded ) sceneSolarSystem.release();
    PostQuitMessage( 0 );
}

//! Decide which scene should be chosen based on the current
//  execution time, and plays/stops music accordingly
void getScene( float *tm ) {
    float t = (*tm);
    sceneCurrent = NULL;
    if( t >= 0.0f && t < dxTimerBPM( 134.1, 37, 0 ) ) { // --> Abstract
        t -= 0.0f;
        sceneCurrent = &sceneAbstract;
    } else if( t < dxTimerBPM( 134.1, 53, 0 ) ) {  // --> Terrain
        t -= dxTimerBPM( 134.1, 37, 0 );
        sceneCurrent = &sceneTerrain;
    } else if( t < dxTimerBPM( 134.1, 85, 0 ) ) {  // --> Abstract 2
        t -= dxTimerBPM( 134.1, 53, 0 );
        sceneCurrent = &sceneAbstract2;
    } else if( t < dxTimerBPM( 134.1, 101, 0 ) ) {  // --> Head
        t -= dxTimerBPM( 134.1, 85, 0 );
        sceneCurrent = &sceneHuman;
    } else if( t < dxTimerBPM( 134.1, 109, 0 ) ) {  // --> Car
        t -= dxTimerBPM( 134.1, 101, 0 );
        sceneCurrent = &sceneCar;
    } else if( t < dxTimerBPM( 134.1, 117, 0 ) ) {  // --> Tunnel
        t -= dxTimerBPM( 134.1, 109, 0 );
        sceneCurrent = &sceneCubeTunnel;
    } else if( t < dxTimerBPM( 134.1, 134, 0 ) ) {  // --> Chrono
        t -= dxTimerBPM( 134.1, 117, 0 );
        sceneCurrent = &sceneChrono;
    } else if( t < dxTimerBPM( 134.1, 134 + 152, 0 ) ) {
        if(!spaceLoaded) {
            loadSpaceScene();
            t = 0.0f;
            execTime = dxTimerBPM( 134.1, 134, 0 );
        } else {
            t -= dxTimerBPM( 134.1, 134, 0 );
        }
        sceneCurrent = &sceneSolarSystem;
    } else {
        sceneCurrent = NULL;
        endDemo();
    }
    (*tm) = t;
    if( sceneCurrent != &sceneSolarSystem && musicState != 1 ) {
        musicState = 1;
        musicStream.stop();
        musicStream.play();
        musicStream.seek( execTime  * 1000.0 );
    }
    if( sceneCurrent == &sceneSolarSystem && musicState != 2 ) {
        musicState = 2;
        musicStream2.stop();
        musicStream2.play();
        musicStream2.seek( t  * 1000.0 );
    }
}

//! Updates and renders the current scene
void renderFrameScene() {
    D3DXMATRIX matView;
    float FOV = 95, nearPlane = 0.1f, farPlane = 5000.0f;
    if ( sceneCurrent == &sceneSolarSystem ) {nearPlane = 0.01f; farPlane = 300000.0f;}

    dxCameraOrbital( &cameraLook, &cameraUp, cameraAngle1, cameraAngle2, cameraDistance );
    cameraPos = freeViewCam;
    cameraTarget = cameraPos + cameraLook;
    shaderSem.cameraPos = cameraPos;

    if( !editMode ) cameraControlMode = CAMERACONTROL_DEMO;

    // Automatically track path
    if( (cameraControlMode == CAMERACONTROL_FREELOOK || cameraControlMode == CAMERACONTROL_FREEORBIT) && trackPathMode) {
        static float trackPathTimer = 0.0f; trackPathTimer -= eulerIntegrate( 1.0f );
        if( trackPathTimer < 0.0f ) {
            wInfo( "> Tracking camera frame %d...\n", cameraPathEditor.numFrames() );
            trackPathTimer = 200.0f;
            cameraPathEditor.addFrame( cameraPos, cameraTarget, cameraUp );
            if( sceneCurrent ) { sceneCurrent->addParamVector(&paramPathEditor); }
        }
    } else trackPathMode = false;

    // Control camera
    D3DXVECTOR3 cameraForward, cameraLeft;
    float cameraDisplacement = eulerIntegrate( cameraDisplacementSpeed );
    switch( cameraControlMode ) {
        case CAMERACONTROL_FREELOOK:
            cameraDistance = 1.0f;
            cameraAngle2+=key[VK_UP]?0.3f:0; cameraAngle1-=key[VK_LEFT]?0.3f:0; cameraAngle2-=key[VK_DOWN]?0.3f:0; cameraAngle1+=key[VK_RIGHT]?0.3f:0;
            D3DXVec3Normalize( &cameraForward, &cameraLook ); 
            D3DXVec3Cross( &cameraLeft, &cameraForward, &cameraUp ); D3DXVec3Normalize( &cameraLeft, &cameraLeft ); 
            cameraForward *= -cameraDisplacement; cameraLeft *= -cameraDisplacement;
            if(key['S']) freeViewCam+=cameraForward; if(key['W']) freeViewCam-=cameraForward; 
            if(key['Q']) freeViewCam.y+=cameraDisplacement; if(key['E']) freeViewCam.y-=cameraDisplacement;
            if(key['D']) freeViewCam+=cameraLeft; if(key['A']) freeViewCam-=cameraLeft;
            break;
        case CAMERACONTROL_FREEORBIT:
            if(key['S']) freeViewCam.x += cameraDisplacement; if(key['W']) freeViewCam.x -= cameraDisplacement;
            if(key['Q']) freeViewCam.y += cameraDisplacement; if(key['E']) freeViewCam.y -= cameraDisplacement;
            if(key['D']) freeViewCam.z += cameraDisplacement; if(key['A']) freeViewCam.z -= cameraDisplacement;
            cameraTarget = freeViewCam;
            cameraPos = freeViewCam + cameraLook;
            shaderSem.cameraPos = cameraPos;
            break;
    }

    // Set camera
    D3DXMatrixLookAtLH(&matView, &cameraPos, &cameraTarget, &cameraUp); 
    d3dDevice->SetTransform(D3DTS_VIEW, &matView);

    // Set default scene for edit mode
    sceneCurrent = &sceneCar;

    sceneCurrent->debugMode = true;
    sceneCurrent->time = execTime;
    FOV = sceneCurrent->FOV;
    postWhite = 0.0f; postBlack = 1.0f;
    // Automatic camera mode
    if( cameraControlMode == CAMERACONTROL_PATH ) {
        cameraPathEditor.setCamera( execTime * 0.1, matView, &shaderSem );
        sceneCurrent->updateParamVector( paramPathEditor, execTime * 0.1 );
    }
    // Demo running camera mode
    if( cameraControlMode == CAMERACONTROL_DEMO ) {
        // Decide what the current scene should be
        float sceneTime = execTime;
        getScene( &sceneTime );
        if( !sceneCurrent ) return;
        if( !sceneCurrent->isLoaded ) return;
        sceneCurrent->time = sceneTime;
        // Update parameters based on new chosen scene
        FOV = sceneCurrent->FOV;
        sceneCurrent->debugMode = false;
        sceneCurrent->update( matView, &shaderSem );
        // Update post processing fader values
        postWhite = whiteFader[execTime]; 
        postBlack = blackFader[execTime];
        postNoise = noiseFader[execTime];
        postDispersion = dispersionFader[execTime];
        postBlurAmount2 = blurFader[execTime];
    }

    if( !sceneCurrent->isLoaded ) return;

    // Update post processing parameters from current scene
    postBloomAmount = sceneCurrent->bloomAmount;
    postBloomThreshold = sceneCurrent->bloomThreshold;
    postBlurAmount = sceneCurrent->blurAmount;
    postSaturation = sceneCurrent->saturate;

    // Set projection matrix
    D3DXMATRIX matProjection;
    D3DXMatrixPerspectiveFovLH(&matProjection, D3DXToRadian(FOV),
                               (FLOAT)d3dWindowWidth / (FLOAT)d3dWindowHeight,
                               nearPlane, farPlane);

    d3dDevice->SetTransform(D3DTS_PROJECTION, &matProjection); 
    D3DXMATRIX matViewProjection = (matView * matProjection);

    // Tell the current scene to do its rendering
    sceneCurrent->render( shaderSem, matView, matProjection );
}

//! Renders a screen buffer, optionally applying an effect to it
void renderScreenBufferWithEffect( RenderPass &renderPass, Shader *shader = NULL, const char *techniqueName = NULL, bool screenTexture = true) {
    if( !shader || !techniqueName)  {
        // Render with no effect
        dxLock();
            dxTexture( 0, renderPass[0] );
            renderPass.render();
        dxUnLock();
        return;
    }
    // Render with effect
    dxLock();
        shader->e()->SetFloat("screenWidth", d3dWindowWidth);
        shader->e()->SetFloat("screenHeight", d3dWindowHeight);
        if( screenTexture ) {
            shader->e()->SetTexture( "screenTexture",  renderPass[0] ); 
            dxTexture( 0, renderPass[0] );
        }
        shader->setGlobalSemantics( &shaderSemDefault );
        for(int a = shader->begin( techniqueName ); a < shader->nPasses(); a++) {
            shader->pass(a);
                renderPass.render();
            shader->endPass();
        }
        shader->end();
    dxUnLock();
}

//! Renders a frame to screen
void renderFrame() {
    // -------> Render scene

    screenHDRBuffer.setAsTarget();
    dxLock();
        renderFrameScene();
    dxUnLock();
    screenHDRBuffer.restoreTarget();
    
    // -------> Render postprocessing effects

    // --> Apply blur passes
    blurPostProcess.e()->SetFloat( "gaussianWidth", 1.0f );
    screenBlurBufferU.setAsTarget();
    renderScreenBufferWithEffect( screenHDRBuffer, &blurPostProcess, "postDiffusionU" );
    screenBlurBufferU.restoreTarget();
    screenBlurBufferV.setAsTarget();
    renderScreenBufferWithEffect( screenBlurBufferU, &blurPostProcess, "postDiffusionV" );
    screenBlurBufferV.restoreTarget();
    blurPostProcess.e()->SetFloat( "gaussianWidth", 2.0f );
    screenBlurBufferU.setAsTarget();
    renderScreenBufferWithEffect( screenBlurBufferV, &blurPostProcess, "postDiffusionU" );
    screenBlurBufferU.restoreTarget();
    screenBlurBufferV.setAsTarget();
    renderScreenBufferWithEffect( screenBlurBufferU, &blurPostProcess, "postDiffusionV" );
    screenBlurBufferV.restoreTarget();
    blurPostProcess.e()->SetFloat( "gaussianWidth", 4.0f );
    screenBlurBufferU.setAsTarget();
    renderScreenBufferWithEffect( screenBlurBufferV, &blurPostProcess, "postDiffusionU" );
    screenBlurBufferU.restoreTarget();
    screenBlurBufferV.setAsTarget();
    renderScreenBufferWithEffect( screenBlurBufferU, &blurPostProcess, "postDiffusionV" );
    screenBlurBufferV.restoreTarget();

    // --> Apply colour adjustment
    adjustmentPostProcess.e()->SetFloat( "postSaturation", postSaturation );
    adjustmentPostProcess.e()->SetFloat( "postContrast", postContrast );
    adjustmentPostProcess.e()->SetFloat( "postExposure", postExposure );
    adjustmentPostProcess.e()->SetFloat( "postBlurAmount", min( postBlurAmount + postBlurAmount2, 1.0f ) );
    adjustmentPostProcess.e()->SetFloat( "postBloomAmount", postBloomAmount );
    adjustmentPostProcess.e()->SetFloat( "postBloomThreshold", postBloomThreshold );
    adjustmentPostProcess.e()->SetFloat( "postWhite", postWhite );
    adjustmentPostProcess.e()->SetFloat( "postBlack", postBlack );
    adjustmentPostProcess.e()->SetFloat( "postNoise", postNoise );
    adjustmentPostProcess.e()->SetFloat( "postDispersion", postDispersion );
    adjustmentPostProcess.e()->SetTexture( "blurTextureU",  screenBlurBufferU[0] ); 
    adjustmentPostProcess.e()->SetTexture( "blurTexture",  screenBlurBufferV[0] ); 
    screenRGBBuffer.setAsTarget();
    renderScreenBufferWithEffect( screenHDRBuffer, &adjustmentPostProcess, "postColourAdjustment" );
    screenRGBBuffer.restoreTarget();

    // --> Apply MLAA
    screenRGBBuffer2.setAsTarget();
    if( !MLAAEnabled ) renderScreenBufferWithEffect( screenRGBBuffer );
    else {
        D3DXVECTOR4 onePixelStepFactor = D3DXVECTOR4(
            1.0f / ((float)(d3dWindowWidth)),
            1.0f / ((float)(d3dWindowWidth)),
            1.0f / ((float)(d3dWindowHeight)),
            1.0f / ((float)(d3dWindowHeight))
        );
        MLAAPostProcess.e()->SetVector( "onePixelStepFactor", &onePixelStepFactor );
        // MLAA Edge Detection
        screenMLAA0Buffer.setAsTarget();
        renderScreenBufferWithEffect( screenRGBBuffer, &MLAAPostProcess, "MLAAEdgeDetection" );
        screenMLAA0Buffer.restoreTarget();
        // MLAA Distance Propagation 0
        screenMLAA1Buffer.setAsTarget();
        MLAAPostProcess.e()->SetFloat( "currentDistanceStep", 0.0f );
        MLAAPostProcess.e()->SetTexture( "distTexture",  screenMLAA0Buffer[0] ); 
        renderScreenBufferWithEffect( screenRGBBuffer, &MLAAPostProcess, "MLAADistancePropagation", false );
        screenMLAA1Buffer.restoreTarget();
        // MLAA Distance Propagation 4
        screenMLAA0Buffer.setAsTarget();
        MLAAPostProcess.e()->SetFloat( "currentDistanceStep", 4.0f );
        MLAAPostProcess.e()->SetTexture( "distTexture",  screenMLAA1Buffer[0] ); 
        renderScreenBufferWithEffect( screenRGBBuffer, &MLAAPostProcess, "MLAADistancePropagation", false );
        screenMLAA0Buffer.restoreTarget();
        // MLAA Distance Propagation 16
        screenMLAA1Buffer.setAsTarget();
        MLAAPostProcess.e()->SetFloat( "currentDistanceStep", 16.0f );
        MLAAPostProcess.e()->SetTexture( "distTexture",  screenMLAA0Buffer[0] ); 
        renderScreenBufferWithEffect( screenRGBBuffer, &MLAAPostProcess, "MLAADistancePropagation", false );
        screenMLAA1Buffer.restoreTarget();
        // MLAA Distance Propagation 64
        screenMLAA0Buffer.setAsTarget();
        MLAAPostProcess.e()->SetFloat( "currentDistanceStep", 64.0f );
        MLAAPostProcess.e()->SetTexture( "distTexture",  screenMLAA1Buffer[0] ); 
        renderScreenBufferWithEffect( screenRGBBuffer, &MLAAPostProcess, "MLAADistancePropagation", false );
        screenMLAA0Buffer.restoreTarget();
        // MLAA Blend Calculation
        screenMLAA1Buffer.setAsTarget();
        MLAAPostProcess.e()->SetTexture( "distTexture",  screenMLAA0Buffer[0] ); 
        MLAAPostProcess.e()->SetTexture( "areaTableTexture",  texture[2] ); 
        renderScreenBufferWithEffect( screenRGBBuffer, &MLAAPostProcess, "MLAABlend", false  );
        screenMLAA1Buffer.restoreTarget();
        // MLAA Final Gather
        MLAAPostProcess.e()->SetTexture( "blendTexture",  screenMLAA1Buffer[0] ); 
        renderScreenBufferWithEffect( screenRGBBuffer, &MLAAPostProcess, "MLAABlur"  );
    }
    screenRGBBuffer2.restoreTarget();

    // --> Apply film grain noise & chromatic dispersion
    adjustmentPostProcess.e()->SetTexture( "noiseTexture",  texture[3] ); 
    adjustmentPostProcess.e()->SetFloat( "postNoiseTime", execTime );
    renderScreenBufferWithEffect( screenRGBBuffer2, &adjustmentPostProcess, "postNoiseDistort" );
    
    // --> Render FPS counter text
    if( showFPS ) {
        dxPrintf( d3dWindowWidth - 110, 10, 0xffffffff, "%.1lf FPS", frameRate);
        dxPrintf( d3dWindowWidth - 110, 35, 0xffffffff, "%.1lf Time", execTime);
    }

    // --> Log FPS to file
    if( logToFile ) {
        logFileTimer -= eulerIntegrate( 1.0f );
        if( logFileTimer < 0.0f ) {
            logFileTimer = 1000.0f;
            logCurrentFPS();
        }
    }
}

