/** \file SceneAbstract.cpp
<h2> Abstract Scenes </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief Abstract Scene
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../GameHeader/SceneAbstract.h"

D3DXVECTOR3 sceneAbstractRandomVector() {
    D3DXVECTOR3 v = D3DXVECTOR3(
        _randfloat( -1.0f, 1.0f ),
        _randfloat( -1.0f, 1.0f ),
        _randfloat( -1.0f, 1.0f )
    );
    D3DXVec3Normalize( &v, &v );
    return v;
}

float SceneAbstract::bezierFunction( int i, double t ) {
    // The basis function for a Bezier spline
    switch (i) {
        case 0: return (1.0-t)*(1.0-t)*(1.0-t);
        case 1: return 3.0*t*(1.0-t)*(1.0-t);
        case 2: return 3.0*t*t*(1.0-t);
        case 3: return t*t*t;
    }
    return 0.0;
}

float SceneAbstract::bezierFunctionDerivative( int i, double t ) {
    // The derivative of the basis function for a Bezier spline
    switch (i) {
        case 0: return (1.0-t)*(1.0-t)*-3.0;
        case 1: return 3.0*t * (-2.0*(1.0-t)) + (1.0-t)*(1.0-t) * 3.0;
        case 2: return -3.0*t*t + (1.0-t)*6.0*t;
        case 3: return 3.0*t*t;
    }
    return 0.0;
}

D3DXVECTOR3 SceneAbstract::evaluateBezier( int i, double t, vector<D3DXVECTOR3> &pts ) {
    D3DXVECTOR3 v( 0.0f, 0.0f, 0.0f );
    for( int j = 0; j < 4; j++ ) {
        if( i + j >= pts.size() ) return v;
        D3DXVECTOR3 p = pts[i + j];
        float bezier = bezierFunction( j, t );
        v.x += bezier * p.x;
        v.y += bezier * p.y;
        v.z += bezier * p.z;
    }
    return v;
}

D3DXVECTOR3 SceneAbstract::evaluateBezierDerivative( int i, double t, vector<D3DXVECTOR3> &pts ) {
    D3DXVECTOR3 v( 0.0f, 0.0f, 0.0f );
    for( int j = 0; j < 4; j++ ) {
        if( i + j >= pts.size() ) return v;
        D3DXVECTOR3 p = pts[i + j];
        float bezier = bezierFunctionDerivative( j, t );
        v.x += bezier * p.x;
        v.y += bezier * p.y;
        v.z += bezier * p.z;
    }
    return v;
}

void SceneAbstract::load() {
    FOV = 65;
    loadSceneAnimationPaths( "Abstract/anim.path" );
    // --> Curve scene
    lineShader.loadFile("Abstract/lineShader.fx");
    textures.load(0, "Abstract/line.png");

    _srand(31337);

    for(int a = 0; a < SCENE_ABS_CURVES; a++) {
        D3DXVECTOR3 dir = sceneAbstractRandomVector();
        for(int b = 0; b < 6; b++) {
            D3DXVECTOR3 disp = sceneAbstractRandomVector();
            D3DXVECTOR3 pos = (dir * b * 1.0f) + (disp * b * 0.3f);
            D3DXVECTOR3 tangent = sceneAbstractRandomVector() * 0.7f + dir;
            D3DXVec3Normalize( &tangent, &tangent );
            tangent *= _randfloat( 0.5f, 0.6f );
            if( b == 0 ) {
                bzPoints[a].push_back( pos );
                bzPoints[a].push_back( pos + tangent );
            } else {
                bzPoints[a].push_back( pos - tangent );
                bzPoints[a].push_back( pos );
            }
            if( b != 0 && b != 6) {
                bzPoints[a].push_back( pos + tangent );
            }
        }
    }
    // --> Fill scene
    fillShader.loadFile("Abstract/fillShader.fx");
    textures.load(1, "Abstract/abstract__04_by_truod05-d39fdni.jpg");
    // --> Ion scene
    ionModel.load("Abstract/ion.obj");
    ionConduitModel.load("Abstract/ionConduit.obj");
    ionShader.loadFile("Abstract/ionShader.fx");
    textures.load(2, "Abstract/ionOcclusion.png");
    textures.load(3, "Abstract/tex.jpg");
    textures.load(4, "Abstract/tex2.jpg");
    textures.load(5, "Abstract/ionOcclusion.png");
    ionBeam = dxPrimitiveCreateTube( 5, 500 );
    // --> City scene
    cityModel.load("Abstract/cityModel.obj");
    cityStreetModel.load("Abstract/cityStreets.obj");
    textures.load(6, "Abstract/cityTexture.png");
    textures.load(7, "Abstract/cityGround.jpg");
    textures.load(8, "Abstract/cityBD.jpg");
    textures.load(9, "Abstract/cityTextureLights.png");
    textures.load(10,"Abstract/cityWater.png");
    textures.load(11,"Abstract/cityBDWater.jpg");
    cityShader.loadFile("Abstract/cityShader.fx");
    D3DXCreateCubeTextureFromFile( d3dDevice, "Abstract/skybox.dds", &cityCubeMap );
    citySkyDome = dxPrimitiveCreateSphereMesh(256, 8, 8);
    // --> Lucy scene
    lucyModel.load("Abstract/lucy.obj");
    lucyShader.loadFile("Abstract/lucyShader.fx");
    D3DXCreateCubeTextureFromFile( d3dDevice, "Abstract/lucyIrradiance.dds", &lucyIrradianceMap );
    D3DXCreateCubeTextureFromFile( d3dDevice, "Abstract/lucyReflection.dds", &lucyReflectionMap );
    // --> Aneurism scene
    aneurismShader.loadFile("Abstract/aneurismShader.fx");
    aneurismSphere = dxPrimitiveCreateSphereMesh( 1.0, 24, 24 );
    D3DXCreateVolumeTextureFromFile( d3dDevice, "Abstract/aneurism.dds", &aneurismVolume );

    isLoaded = true;
}

void SceneAbstract::renderCurves(ShaderGlobalSemantics &shaderSem, 
        D3DXMATRIX &matView, D3DXMATRIX &matProjection) {

    d3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);

    bloomAmount = 0.7f; bloomThreshold = 0.5f;

    shaderSem.matView = matView;
    shaderSem.matProj = matProjection;

    vector<SceneAbstractLineSegment> ZSorter;

    for(int k = 0; k < SCENE_ABS_CURVES; k++) {
        for(int i = 0; i < ((int)bzPoints[k].size()) - 3; i+=3) {
            for(int j = 0; j < SCENE_ABS_BEZIERSTEPS; j++) {
                // Calcualte bezier curve
                shaderSem.matWorld = dxGetMatrix();
                float stepLength1  = (float)j / (float)SCENE_ABS_BEZIERSTEPS;
                float stepLength2 = (float)(j + 1) / (float)SCENE_ABS_BEZIERSTEPS;
                SceneAbstractLineSegment seg;
                seg.p1  = evaluateBezier( i, stepLength1, bzPoints[k] );
                seg.d1  = evaluateBezierDerivative( i, stepLength1, bzPoints[k] );
                seg.p2  = evaluateBezier( i, stepLength2, bzPoints[k] );
                seg.d2  = evaluateBezierDerivative( i, stepLength2, bzPoints[k] );
                D3DXVECTOR4 transformedVertex;
                D3DXVec3Transform(&transformedVertex, &((seg.p1+seg.p2)*0.5), 
                    &(shaderSem.matWorld * shaderSem.matView * shaderSem.matProj));
                transformedVertex /= transformedVertex.w;
                seg.z = transformedVertex.z;
                ZSorter.push_back( seg );
                
            }
        }
    }

    //sort( ZSorter.begin(), ZSorter.end() );

    static float time = 0.0f;
    time += 0.01f;

    dxPushMatrix();
        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        lineShader.setGlobalSemantics( &shaderSem );
        lineShader.e()->SetTexture("colourTexture", textures[0]); dxTexture(0, textures[0]);
        lineShader.e()->SetFloat("time", time);
        for(int a = 0; a < ZSorter.size(); a++) {
            lineShader.e()->SetVector( "point0", &((D3DXVECTOR4)(ZSorter[a].p1)) );
            lineShader.e()->SetVector( "point1", &((D3DXVECTOR4)(ZSorter[a].p2)) );
            lineShader.e()->SetVector( "tangent0", &((D3DXVECTOR4)(ZSorter[a].d1)) );
            lineShader.e()->SetVector( "tangent1", &((D3DXVECTOR4)(ZSorter[a].d2)) );
            for(int a = lineShader.begin("Line"); a < lineShader.nPasses(); a++) {
                lineShader.pass(a);
                dxFastQuad();
                lineShader.endPass();
            }
            lineShader.end();
        }
    dxPopMatrix();

}

void SceneAbstract::renderFill(ShaderGlobalSemantics &shaderSem, 
        D3DXMATRIX &matView, D3DXMATRIX &matProjection) {

    d3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);

    bloomAmount = 0.5f; bloomThreshold = 0.7f;

    float quadSize = 8.0f;
    float fillTime = debugMode?(time * 10.0f):(animFrameTime * 7.5f);
    shaderSem.matView = matView;
    shaderSem.matProj = matProjection;
    dxRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
    dxPushMatrix();
        dxRotateY( 90.0f );
        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        fillShader.setGlobalSemantics( &shaderSem );
        fillShader.e()->SetTexture("colourTexture", textures[1]); dxTexture(0, textures[1]);
        fillShader.e()->SetFloat("time", fillTime);
        for(int a = fillShader.begin("Fill"); a < fillShader.nPasses(); a++) {
            fillShader.pass(a);
            dxQuad( quadSize, quadSize, -quadSize*0.5, -quadSize*0.5, 0xffffffff, false );
            fillShader.endPass();
        }
        fillShader.end();
    dxPopMatrix();
    dxRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
}

void SceneAbstract::renderIon(ShaderGlobalSemantics &shaderSem, 
        D3DXMATRIX &matView, D3DXMATRIX &matProjection) {

    d3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(61, 51, 139), 1.0f, 0);

    bloomAmount = 0.2f; bloomThreshold = 1.1f;

    float ionTime = debugMode?(time * 10.0f * ionTimeScale):(animFrameTime * 10.0f * ionTimeScale);

    float lightRingPeriod = 50.0f;
    float lightRingMaxRadius = 250.0f;
    float lightRingRadius = (fmod( ionTime, lightRingPeriod ) / lightRingPeriod) * lightRingMaxRadius;
    float lightRingAlpha = 1.0f - (lightRingRadius / lightRingMaxRadius);

    shaderSem.matView = matView;
    shaderSem.matProj = matProjection;

    dxPushMatrix();
        dxScale( 0.1f, 0.1f, 0.1f );
        dxTranslate(0.0f, -100.0f, 0.0f);
        dxCommitMatrix();
        dxTexture(0, textures[4]);

        dxPushMatrix();
            ionShader.e()->SetTexture("colourTexture", textures[4]); dxTexture(0, textures[4]);
            ionShader.e()->SetTexture("AOTexture", textures[5]); dxTexture(1, textures[5]);
            ionModel.render(&ionShader, &shaderSem, "Generator");
        dxPopMatrix();

        dxPushMatrix();
            dxRotateY( ionTime * 18.0 );
            dxCommitMatrix();
            ionShader.e()->SetTexture("colourTexture", textures[4]); dxTexture(0, textures[4]);
            ionShader.e()->SetTexture("AOTexture", textures[5]); dxTexture(1, textures[5]);
            ionConduitModel.render(&ionShader, &shaderSem, "Generator");
        dxPopMatrix();
        
        dxRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
        dxPushMatrix();
            dxRotateX( 90 );
            shaderSem.matWorld = dxGetMatrix();
            dxCommitMatrix();
            ionShader.setGlobalSemantics( &shaderSem );
            ionShader.e()->SetTexture("colourTexture", textures[3]); dxTexture(0, textures[3]);
            ionShader.e()->SetFloat("lightRadius", lightRingRadius);
                ionShader.e()->SetFloat("lightAlpha", lightRingAlpha);
            for(int a = ionShader.begin("Ground"); a < ionShader.nPasses(); a++) {
                ionShader.pass(a);
                dxQuad( 10000, 10000, -5000, -5000 );
                ionShader.endPass();
            }
            ionShader.end();
            dxTranslate(0, 0.0f, -465.0f);
            shaderSem.matWorld = dxGetMatrix();
            ionShader.setGlobalSemantics( &shaderSem );
            dxCommitMatrix();
            for(int a = ionShader.begin("Ground"); a < ionShader.nPasses(); a++) {
                ionShader.pass(a);
                dxQuad( 10000, 10000, -5000, -5000 );
                ionShader.endPass();
            }
            ionShader.end();

        dxPopMatrix();
        dxRenderState(D3DRS_CULLMODE, D3DCULL_CCW);

        dxPushMatrix();
            dxRotateX( 90 );
            dxTranslate(0.0, 0.0, -230);
            for(int k = 0; k < 8; k++) {
                dxPushMatrix();
                    dxScale( 1.0f + k * 0.25f,
                             1.0f + k * 0.25f,
                             1.0f + k * 0.25f );
                    shaderSem.matWorld = dxGetMatrix();
                    dxCommitMatrix();
                    ionShader.setGlobalSemantics( &shaderSem );
                    for(int a = ionShader.begin("Beam"); a < ionShader.nPasses(); a++) {
                        ionShader.pass(a);
                        dxTube( ionBeam );
                        ionShader.endPass();
                    }
                    ionShader.end();
                dxPopMatrix();
            }
        dxPopMatrix();

    dxPopMatrix();
}

void SceneAbstract::renderCity(ShaderGlobalSemantics &shaderSem, 
        D3DXMATRIX &matView, D3DXMATRIX &matProjection) {
    
    float timeCity = time * 5.0f;

    bloomAmount = 0.57f; bloomThreshold = 0.8f;

    shaderSem.matView = matView;
    shaderSem.matProj = matProjection;

    dxPushMatrix();
        dxScale( 10, 10,10 );
        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        cityShader.setGlobalSemantics( &shaderSem );
        cityShader.e()->SetTexture("cubeMapTexture", cityCubeMap); 
        for(int a = cityShader.begin("Skybox"); a < cityShader.nPasses(); a++) {
            cityShader.pass(a);
                dxSphere( citySkyDome );
            cityShader.endPass();
        }
        cityShader.end();
    dxPopMatrix();


    dxPushMatrix();
        dxTranslate(10.0f, 0.0f, 0.0f);
        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        cityShader.setGlobalSemantics( &shaderSem );
        cityShader.e()->SetTexture("cityTexture", textures[6]); dxTexture(0, textures[6]);
        cityShader.e()->SetTexture("lightsTexture", textures[9]); dxTexture(1, textures[9]);
        for(int a = cityShader.begin("Building"); a < cityShader.nPasses(); a++) {
            cityShader.pass(a);
                cityModel.render();
            cityShader.endPass();
        }
        cityShader.end();
        //cityStreetModel.render();
    dxPopMatrix();

    dxPushMatrix();
        dxRotateX( -90 );

        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        cityShader.setGlobalSemantics( &shaderSem );
        cityShader.e()->SetTexture("colourTexture", textures[7]); dxTexture( 0, textures[7] );
        cityShader.e()->SetTexture("specTexture", textures[10]); dxTexture( 1, textures[10] );
        cityShader.e()->SetFloat("worldPosYoffset", 0.0f);
        cityShader.e()->SetFloat("time", timeCity);
        for(int a = cityShader.begin("Ground"); a < cityShader.nPasses(); a++) {
            cityShader.pass(a);
                dxQuad( 200, 200, -100, -100);
            cityShader.endPass();
        }
        cityShader.end();

        dxTranslate(0.0f, 0.0f, -1.0f);
        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        cityShader.setGlobalSemantics( &shaderSem );
        cityShader.e()->SetTexture("colourTexture", textures[8]); dxTexture( 0, textures[8] );
        cityShader.e()->SetTexture("specTexture", textures[11]); dxTexture( 1, textures[11] );
        cityShader.e()->SetFloat("worldPosYoffset", 1.0f);
        for(int a = cityShader.begin("Ground"); a < cityShader.nPasses(); a++) {
            cityShader.pass(a);
                dxQuad( 2000, 2000, -1000, -1000);
            cityShader.endPass();
        }
        cityShader.end();
    dxPopMatrix();

}

void SceneAbstract::renderLucy(ShaderGlobalSemantics &shaderSem, 
        D3DXMATRIX &matView, D3DXMATRIX &matProjection) {

    shaderSem.matView = matView;
    shaderSem.matProj = matProjection;

    bloomAmount = 0.3f; bloomThreshold = 1.1f;

    dxPushMatrix();
        lucyShader.e()->SetTexture("irradianceMapTexture", lucyIrradianceMap); 
        lucyShader.e()->SetTexture("reflectionMapTexture", lucyReflectionMap); 
        lucyModel.render(&lucyShader, &shaderSem, "Statue");
    dxPopMatrix();

    dxPushMatrix();
        dxScale( 10, 10, 10 );
        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        lucyShader.setGlobalSemantics( &shaderSem );
        for(int a = lucyShader.begin("Skybox"); a < lucyShader.nPasses(); a++) {
            lucyShader.pass(a);
            dxSphere( citySkyDome );
            lucyShader.endPass();
        }
        lucyShader.end();
    dxPopMatrix();

}

void SceneAbstract::renderAneurism(ShaderGlobalSemantics &shaderSem, 
        D3DXMATRIX &matView, D3DXMATRIX &matProjection) {
    shaderSem.matView = matView;
    shaderSem.matProj = matProjection;
    bloomAmount = 0.6f; bloomThreshold = 0.65f;
    d3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 0), 1.0f, 0);
    dxPushMatrix();
        shaderSem.matWorld = dxGetMatrix();
        aneurismShader.setGlobalSemantics( &shaderSem );
        dxCommitMatrix();
        aneurismShader.e()->SetTexture( "volumeTexture", aneurismVolume );
        for(int a = aneurismShader.begin("Aneurism"); a < aneurismShader.nPasses(); a++) {
            aneurismShader.pass(a);
            dxSphere( aneurismSphere );
            aneurismShader.endPass();
        }
        aneurismShader.end();
    dxPopMatrix();
}

void SceneAbstract::render(ShaderGlobalSemantics &shaderSem, 
        D3DXMATRIX &matView, D3DXMATRIX &matProjection) {
    if( debugMode ) {
        FOV = 65;
        renderCity( shaderSem, matView, matProjection ); 
        return;
    }
    bloomAmount = 0.5f; bloomThreshold = 0.7f;
    switch( animFrame ) {
        case 0: case 1:  case 2:
            renderCity( shaderSem, matView, matProjection ); 
            break;
        case 3: case 4: case 5: case 6:
            renderLucy( shaderSem, matView, matProjection ); 
            break;
        case 7: 
            renderCity( shaderSem, matView, matProjection ); 
            break;
        case 8:
            renderCurves( shaderSem, matView, matProjection ); 
            break;
        case 9: case 10:
            renderAneurism( shaderSem, matView, matProjection ); 
            break;
        case 11:case 12:case 13:case 14:case 15:case 16:
            ionTimeScale = animFrame>=13 ? 3.5f : 1.0f;
            renderIon( shaderSem, matView, matProjection ); 
            break;
        case 17: case 18:
            renderCurves( shaderSem, matView, matProjection ); 
            break;
        case 19:
            renderFill( shaderSem, matView, matProjection ); 
            break;
    }
}

void SceneAbstract::update( D3DXMATRIX &matView, ShaderGlobalSemantics *shaderSem ) {
    getSceneAnimationFrame( &animFrame, &animFrameTime, &animFrameLength );
    if( animFrameLength < 0.000001 ) animFrameLength = 0.000001;
    FOV = 90;
    double cameraPathTimeFactor = animFrameTime / animFrameLength;
    blurAmount = 0.0f; saturate = 1.0f;
    switch( animFrame ) {
        case 0: blurAmount = 1.0f; saturate = 0.2f; FOV = 90; break;
            case 1:  FOV = 90; break; case 2: FOV = 65; break;
        case 3: if( cameraPathTimeFactor < 0.5f ) {blurAmount = 1.0f;saturate = 0.8f;} FOV = 95; break;
            case 4: case 5: FOV = 65; break;
        case 6: FOV = 65; break;
        case 7: FOV = 110; break;
        case 8: FOV = 125; break;
        case 9: FOV = 125; cameraPathTimeFactor = 1.0 - cos( cameraPathTimeFactor * D3DX_PI * 0.5 ); break;
        case 10:FOV = 125; cameraPathTimeFactor = sin( cameraPathTimeFactor * D3DX_PI * 0.5 ); break;
        case 11:case 12:case 13:case 14:case 15:case 16: 
            FOV = 100; saturate = ( animFrame>=15?0.5f:1.0f ); blurAmount = ( animFrame>=15?0.5f:0.0f ); break;
        case 17:case 18:FOV = 120; break;
        case 19:FOV = 120; break;
    }
    cameraPath[ animFrame ].setCamera( cameraPathTimeFactor * SCENE_DEFAULT_CAMERAFACTOR, matView, shaderSem );
    if( debugMode ) FOV = 125;
}

void SceneAbstract::release() {
    textures.clear();
    lineShader.release();
    fillShader.release();
    ionModel.release();
    ionConduitModel.release();
    ionShader.release();
    ionBeam->release();
    cityModel.release();
    cityStreetModel.release();
    cityShader.release();
    cityCubeMap->Release();
    citySkyDome->release();
    lucyModel.release();
    lucyShader.release();
    lucyIrradianceMap->Release();
    lucyReflectionMap->Release();
    aneurismShader.release();
    aneurismSphere->release();
    aneurismVolume->Release();
    isLoaded = false;
}