/** \file SceneHuman.cpp
<h2> NVIDIA Human Head Scene </h2> <br>
Copyright (C) 2011 Xi Chen <br>

-------------------------------------------------------------------------<br>
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version. <br>

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details. <br>

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>. <br>
-------------------------------------------------------------------------<br>

\brief NVIDIA Human Head Scene
\author Xi Chen hypernewbie@gmail.com
\date 15/11/2011

*/

#include "../GameHeader/SceneHuman.h"

void SceneHuman::load() {
    loadSceneAnimationPaths( "Human/anim.path" );
    displayImageDumps = false;
    irradianceTextureSize = SCENEHUMAN_DEFAULT_IRR_TEXSIZE;

    humanSkyDome = dxPrimitiveCreateSphereMesh(256, 8, 8);

    humanHeadModel.load( "Human/nvidiaHumanHead.obj" );
    textures.load(0, "Human/AdrianFinalNormal.tga");
    textures.load(1, "Human/FaceStretchXY.tga");
    textures.load(2, "Human/AdrianAlbedo.tga");
    textures.load(3, "Human/AdrianSpec.tga");
    textures.load(4, "Human/rho_d.tga");

    irradianceShader.loadFile( "Human/irradianceShader.fx" );
    irradianceTexturePass.createTarget( D3DFMT_A8R8G8B8, irradianceTextureSize, 
        irradianceTextureSize );

    diffuseVarianceScale = SCENEHUMAN_DEFAULT_DIFFUSE_VARIANCE_SCALE;
    diffuseGaussianShader.loadFile( "Human/diffusionShader.fx" );
    for( int a = 0; a < SCENEHUMAN_NUM_GAUSSIANS; a++ ) {
        diffusePassU[a].createTarget( D3DFMT_A8R8G8B8, irradianceTextureSize, 
            irradianceTextureSize );
        diffusePassV[a].createTarget( D3DFMT_A8R8G8B8, irradianceTextureSize, 
            irradianceTextureSize );
    }

    skinShader.loadFile( "Human/skinShader.fx" );
    D3DXCreateCubeTextureFromFile( d3dDevice, "Human/cubeIrradianceMap.dds", &skinCubeIrradianceMap );

    headShadowMapShader.loadFile( "Human/shadowShader.fx" );
    headShadowMapSize = SCENEHUMAN_DEFAULT_SHADOWMAP_SIZE;
    headShadowMap.initalise( D3DFMT_A16B16G16R16F, headShadowMapSize, headShadowMapSize, 
        SCENEHUMAN_SHADOWMAP_BOXSIZE, SCENEHUMAN_SHADOWMAP_BOXSIZE);

    headLightPos = D3DXVECTOR3(100000.0 ,50000.0f, 0.0f);
    headAmbientAmount = 0.45f;
    isLoaded = true;
}

void SceneHuman::render(ShaderGlobalSemantics &shaderSem, 
        D3DXMATRIX &matView, D3DXMATRIX &matProjection) {
    if( debugMode ) time = 0;
    FOV = 75;
    D3DXVECTOR4 headLightPosition = (D3DXVECTOR4) headLightPos;
    headLightPosition.y += max(time - 14.0f, 0) * 90000.0f;
    headLightPosition.x -= max(time - 18.0f, 0) * 120000.0f;
    headLightPosition.z += max(time - 20.0f, 0) * 5000.0f;
    D3DXVECTOR3 headLightShadowPosition; D3DXVec3Normalize( &headLightShadowPosition, &((D3DXVECTOR3)(headLightPosition)) );
    headLightShadowPosition *= 111.8;
    float headAmbientLightAmount = max( headAmbientAmount - max(time - 15.0f, 0) * 0.05f, 0.0 );
    float headSkyboxFactor = headAmbientLightAmount / headAmbientAmount;
    
    shaderSem.matView = matView;
    shaderSem.matProj = matProjection;

    bloomAmount = 0.275f; bloomThreshold = 0.55f;

    // -----> Render Skybox
    dxPushMatrix();
        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        skinShader.setGlobalSemantics( &shaderSem );
        skinShader.e()->SetTexture("cubeMapTexture", skinCubeIrradianceMap); 
        skinShader.e()->SetFloat("skyboxFactor", headSkyboxFactor );
        for(int a = skinShader.begin("Skybox"); a < skinShader.nPasses(); a++) {
            skinShader.pass(a);
                dxSphere( humanSkyDome );
            skinShader.endPass();
        }
        skinShader.end();
    dxPopMatrix();

    // -----> Render Shadowmap
    headShadowMap.setLightPos( headLightShadowPosition );
    headShadowMap.setLightTarget( D3DXVECTOR3( 0.0f , 0.0f, 0.0f ) );
    headShadowMap.pushState();
    ShaderGlobalSemantics shadowShaderSem = shaderSem;
    dxPushMatrix();
        shadowShaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        headShadowMapShader.setGlobalSemantics( &shadowShaderSem );
        headShadowMap.setSemantics( &shadowShaderSem );
        humanHeadModel.render( &headShadowMapShader, &shadowShaderSem, "ShadowMap" );
    dxPopMatrix();
    headShadowMap.popState();
    D3DXMATRIX headShadowMapMatrix = headShadowMap.getLightMatrix( &shadowShaderSem );

    // -----> Set up variance fit model
    diffuseVarianceModel[0] = sqrtf( 0.0064f ) * diffuseVarianceScale;
    diffuseVarianceModel[1] = sqrtf( 0.0484f ) * diffuseVarianceScale;
    diffuseVarianceModel[2] = sqrtf( 0.187f ) * diffuseVarianceScale;
    diffuseVarianceModel[3] = sqrtf( 0.567f ) * diffuseVarianceScale;
    diffuseVarianceModel[4] = sqrtf( 1.99f ) * diffuseVarianceScale;
    diffuseVarianceModel[5] = sqrtf( 7.4f ) * diffuseVarianceScale;

    // -----> Render Irradiance texture
    irradianceTexturePass.setAsTarget();
    dxLock();
    dxPushMatrix();

        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        irradianceShader.setGlobalSemantics( &shaderSem );
        irradianceShader.e()->SetTexture("normalTexture", textures[0]); dxTexture(0, textures[0]);
        irradianceShader.e()->SetTexture("diffuseTexture", textures[2]); dxTexture(1, textures[2]);
        irradianceShader.e()->SetTexture("shadowTexture", headShadowMap); dxTexture(2, headShadowMap);
        irradianceShader.e()->SetTexture("cubeMapTexture", skinCubeIrradianceMap);
        irradianceShader.e()->SetMatrix("matShadow", &headShadowMapMatrix);
        irradianceShader.e()->SetVector("lightPos", &headLightPosition);
        irradianceShader.e()->SetFloat("ambientAmount", headAmbientLightAmount );

        humanHeadModel.render(&irradianceShader, &shaderSem, "IrradianceTexture");

    dxPopMatrix();
    dxUnLock();
    irradianceTexturePass.restoreTarget();
    if( displayImageDumps ) irradianceTexturePass.dump( "imageDump/irradianceMap.bmp" );

    // -----> Calculate the separate gaussians
    diffuseGaussianShader.e()->SetTexture("stretchTexture", textures[1]);
    for( int a = 0; a < SCENEHUMAN_NUM_GAUSSIANS; a++ ) {
        // Set radius
        diffuseGaussianShader.e()->SetFloat("gaussianWidth", 
            ( diffuseVarianceModel[a] - (a==0?0.0f:diffuseVarianceModel[a-1]) )
        );
        // Gaussian blur - U direction
        diffusePassU[a].setAsTarget();
        dxLock();
        dxPushMatrix();
            dxCommitMatrix();
            diffuseGaussianShader.e()->SetTexture("irradianceTexture", a==0 ? irradianceTexturePass[0] : diffusePassV[a-1][0] ); 
            dxTexture(0, a==0 ? irradianceTexturePass[0] : diffusePassV[a-1][0]);
            for(int b = diffuseGaussianShader.begin("gaussianDiffusionU"); b < diffuseGaussianShader.nPasses(); b++) {
                diffuseGaussianShader.pass(b);
                dxFastQuad();
                diffuseGaussianShader.endPass();
            }
            diffuseGaussianShader.end();
        dxPopMatrix();
        dxUnLock();
        diffusePassU[a].restoreTarget();
        // Gaussian blur - V direction
        diffusePassV[a].setAsTarget();
        dxLock();
        dxPushMatrix();
           dxCommitMatrix(); 
           diffuseGaussianShader.e()->SetTexture("irradianceTexture", diffusePassU[a][0]);
           dxTexture(0, diffusePassU[a][0]);
           for(int b = diffuseGaussianShader.begin("gaussianDiffusionV"); b < diffuseGaussianShader.nPasses(); b++) {
                diffuseGaussianShader.pass(b);
                dxFastQuad();
                diffuseGaussianShader.endPass();
            }
            diffuseGaussianShader.end();
        dxPopMatrix();
        dxUnLock();
        diffusePassV[a].restoreTarget();
    }
    
    if( displayImageDumps ) {
        diffusePassU[0].dump( "imageDump/diffuseMap0U.bmp" );
        diffusePassV[0].dump( "imageDump/diffuseMap0V.bmp" );
        diffusePassU[1].dump( "imageDump/diffuseMap1U.bmp" );
        diffusePassV[1].dump( "imageDump/diffuseMap1V.bmp" );
        diffusePassU[2].dump( "imageDump/diffuseMap2U.bmp" );
        diffusePassV[2].dump( "imageDump/diffuseMap2V.bmp" );
        diffusePassU[3].dump( "imageDump/diffuseMap3U.bmp" );
        diffusePassV[3].dump( "imageDump/diffuseMap3V.bmp" );
        diffusePassU[4].dump( "imageDump/diffuseMap4U.bmp" );
        diffusePassV[4].dump( "imageDump/diffuseMap4V.bmp" );
        diffusePassU[5].dump( "imageDump/diffuseMap5U.bmp" );
        diffusePassV[5].dump( "imageDump/diffuseMap5V.bmp" );
    }

    // -----> Render skin
    dxPushMatrix();

        shaderSem.matWorld = dxGetMatrix();
        dxCommitMatrix();
        skinShader.setGlobalSemantics( &shaderSem );
        dxTexture(0, textures[2]);
        skinShader.e()->SetTexture("diffuseTexture", textures[2]); 
        skinShader.e()->SetTexture("normalTexture", textures[0]); 
        skinShader.e()->SetTexture("specTexture", textures[3]);
        skinShader.e()->SetTexture("rhoTexture", textures[4]);
        skinShader.e()->SetTexture("irradianceTexture0", diffusePassV[0]); 
        skinShader.e()->SetTexture("irradianceTexture1", diffusePassV[1]); 
        skinShader.e()->SetTexture("irradianceTexture2", diffusePassV[2]); 
        skinShader.e()->SetTexture("irradianceTexture3", diffusePassV[3]); 
        skinShader.e()->SetTexture("irradianceTexture4", diffusePassV[4]); 
        skinShader.e()->SetTexture("irradianceTexture5", diffusePassV[5]); 
        skinShader.e()->SetTexture("cubeMapTexture", skinCubeIrradianceMap);
        skinShader.e()->SetTexture("shadowTexture", headShadowMap);
        skinShader.e()->SetMatrix("matShadow", &headShadowMapMatrix);
        skinShader.e()->SetVector("lightPos", &headLightPosition);
        skinShader.e()->SetFloat("ambientAmount", headAmbientLightAmount );

        humanHeadModel.render(&skinShader, &shaderSem, "Skin");

    dxPopMatrix();
    

}

void SceneHuman::update( D3DXMATRIX &matView, ShaderGlobalSemantics *shaderSem ) {
    getSceneAnimationFrame( &animFrame, &animFrameTime, &animFrameLength );
    if( animFrameLength < 0.000001 ) animFrameLength = 0.000001;
    FOV = 90;
    double cameraPathTimeFactor = animFrameTime / animFrameLength;
    cameraPath[ animFrame ].setCamera( cameraPathTimeFactor * 0.8, matView, shaderSem );
}

void SceneHuman::release() {
    textures.clear();
    humanSkyDome->release();
    humanHeadModel.release();
    irradianceShader.release();
    irradianceTexturePass.releaseTarget();
    diffuseGaussianShader.release();
    for(int a = 0; a < SCENEHUMAN_NUM_GAUSSIANS; a++) {
        diffusePassU[a].releaseTarget();
        diffusePassV[a].releaseTarget();
    }
    skinShader.release();
    skinCubeIrradianceMap->Release();
    headShadowMap.release();
    headShadowMapShader.release();
    isLoaded = false;
}